/*
 * Config.c
 * (c) 2003 G. Forrest Cook, Jonathan Corbet
 * License: GPL Version 2.0
 * Deal with configuration information.
 */
# include <stdlib.h>
# include <string.h>
# include <stdio.h>
# include <ctype.h>
# include <syslog.h>

# include "Config.h"

/*
 * Nowadays we can have more than one config file, so here is the list of
 * them.
 */
typedef struct _cflist
{
	FILE *cl_File;			/* The actual file 	*/
	char *cl_Name;			/* internal name of the file */
	struct _cflist *cl_Next;	/* Next in chain	*/
} cflist;

static cflist *CFList = 0;
static int Griped = 0;		/* Have we moaned about a bad line? */

/*
 * A directory in which we look for config files.
 */
static char CDir[200];

/*
 * Forwards.
 */
static int CF_SplitLine (char *, char **, char **);
static const char *CF_SearchCFile (FILE *, const char *);
static FILE *cf_GetFile (const char *name);





void
cf_ConfigDir (const char *dir)
/*
 * Set the configuration directory.
 */
{
	strcpy (CDir, dir);
}





int
CF_SetConfigFile (const char *file, const char *name)
/*
 * Set the file from which we'll read configuration information.
 */
{
	cflist *ent = (cflist *) malloc (sizeof (cflist));
/*
 * Simply try to open the file.
 */
	if ((ent->cl_File = fopen (file, "r")) == NULL)
	{
		free (ent);
		return (0);
	}
/*
 * Add it to the list.
 */
	ent->cl_Next = 0;
	ent->cl_Name = strdup (name);
	if (! CFList)
		CFList = ent;
	else
	{
		cflist *cp;
		for (cp = CFList; cp->cl_Next; cp = cp->cl_Next)
			;
		cp->cl_Next = ent;
	}
	return (1);
}




void
CF_GetConfigString (const char *param, char *result, const char *deflt)
/*
 * Read a string from the config file, return default if it's not found.
 */
{
	const char *cfp;

        cfp = CF_GetConfigParam (param);

        if (cfp == NULL)
		strcpy (result, deflt);
	else
		strcpy (result, cfp);
}




void
CF_GetConfigInt (const char *param, int *result, int deflt)
/*
 * Read an integer from the config file, return default if it's not found.
 */
{
	const char *cfp;

        cfp = CF_GetConfigParam (param);

        if (cfp == NULL)
		*result = deflt;
	else if (sscanf (cfp, "%d", result) != 1)
		*result = deflt; /* Shouldn't we bitch mightily here?? */
}




float
CF_GetConfigFloat (const char *param, float def)
/*
 * Return a floating-point parameter.  Inconsistent with the interface of
 * the other routines, but I just couldn't replicate that...
 */
{
	const char *cfp = CF_GetConfigParam (param);
	float ret;
	
        if (! cfp || sscanf (cfp, "%f", &ret) != 1)
		return (def);
	return (ret);
}





const char *
CF_GetConfigParam (const char *param)
/*
 * Look up a configuration parameter.  Returns the value of the parameter
 * (good until the next call) if found, NULL otherwise.
 */
{
	cflist *ent;
	const char *ret;

	for (ent = CFList; ent; ent = ent->cl_Next)
		if ((ret = CF_SearchCFile (ent->cl_File, param)))
			return (ret);
	return (NULL);
}





static const char *
CF_SearchCFile (FILE *fp, const char *param)
/*
 * Look for this parameter in a single file.
 */
{
# define MAXLINE 240
	static char line[MAXLINE];
	char *fparam, *value;
/*
 * Start plowing through the file.
 */
	rewind (fp);
	while (fgets (line, MAXLINE, fp))
	{
		if (! CF_SplitLine (line, &fparam, &value))
			continue;
		if (! strcmp (fparam, param))
			return (value);
	}
	return (NULL);
}






static int
CF_SplitLine (char *line, char **param, char **value)
/*
 * Split up a line into parameter and value parts.  Returns nonzero
 * if things worked out.
 */
{
/*
 * Skip any leading white space.
 */
	while (*line != '\0' && isspace(*line))
		line++;
	if (*line == '\0' || *line == '#')
		return (0);
/*
 * This is the beginning of the parameter name.  Skip up to white space
 * or colon.
 */
	*param = line;
	while (*line != '\0' && !isspace (*line) && *line != ':')
		line++;
	if (! *line)
	{
		if (! Griped++)
			syslog (LOG_ERR, "Bad config file line");
		return (0);
	}
	*line++ = '\0';
/*
 * Skip white space once again.
 */
	while (isspace (*line) && *line != '\0')
		line++;
	*value = line;
/*
 * Trim the trailing newline.
 */
	if (line[strlen (line) - 1] == '\n')
		line[strlen (line) - 1] = '\0';
	return (1);
}





/*
 * New interface routines below here.
 */

const char *
cf_GetParam (const char *cfile, const char *param)
/*
 * Get a parameter out of this particular configuration file.
 */
{
	FILE  *fd = cf_GetFile (cfile);

	if (! fd)
		return (0);
	return (CF_SearchCFile (fd, param));
}





static FILE *
cf_GetFile (const char *name)
/*
 * Look up this config file.
 */
{
	cflist *cp;

	for (cp = CFList; cp; cp = cp->cl_Next)
		if (! strcmp (cp->cl_Name, name))
			return (cp->cl_File);
	return (0);
}



int
cf_GetInt (const char *cfile, const char *param, int deflt)
/*
 * Read an integer from the config file, return default if it's not found.
 */
{
	const char *cfp = cf_GetParam (cfile, param);
	int ret;

        if (cfp == NULL || sscanf (cfp, "%d", &ret) != 1)
		return (deflt);
	return (ret);
}




float
cf_GetFloat (const char *cfile, const char *param, float deflt)
/*
 * Read an integer from the config file, return default if it's not found.
 */
{
	const char *cfp = cf_GetParam (cfile, param);
	float ret;

        if (cfp == NULL || sscanf (cfp, "%f", &ret) != 1)
		return (deflt);
	return (ret);
}
