/*
 *-----------------------------------------------------------------------------
 * WsIngest.c  Utilities for Weather Ingesters Apr 28, 1997 [fc]
 * (c) 2003 G. Forrest Cook, Jonathan Corbet
 * License: GPL Version 2.0
 *-----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <signal.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "WsIngest.h"
#include "Config.h"

extern int Shortfing;

#if 0
int tm_sec;                   /* Seconds.     [0-60] (1 leap second) */
int tm_min;                   /* Minutes.     [0-59] */
int tm_hour;                  /* Hours.       [0-23] */
int tm_mday;                  /* Day.         [1-31] */
int tm_mon;                   /* Month.       [0-11] */
int tm_year;                  /* Year - 1900.  */
int tm_wday;                  /* Day of week. [0-6] */
int tm_yday;                  /* Days in year.[0-365] */
int tm_isdst;                 /* DST.         [-1/0/1]*/
#endif

/*
 * Write the gmt time and a string to stdout.
 */
void gmt_log (char *str)
{
	char tstr[30];
	long t = time ((time_t *) 0);
	strcpy (tstr, asctime(gmtime(&t)));
	tstr[strlen(tstr)-1] = '\0';
	printf ("%s: %s\n", tstr, str);
}

/*
 * Return the minute from the gmt time.
 */
int gmt_minute ()
{
	struct tm *gmt;
	long t = time ((time_t *) 0);
	gmt = gmtime(&t);
	return (gmt->tm_min);
}



#ifdef SOLARIS
/*
 * sol_usleep dummy handler
 */
void dummy_hndl ()
{
}



/*
 * Solaris substitute for isleep
 */
void sol_usleep (int usec)
{
        struct itimerval tmrv;

        signal (SIGALRM, &dummy_hndl);
/*
 * Stuff auto reload value and next delay time into itimerval struct.
 */
        tmrv.it_interval.tv_sec = 0;
        tmrv.it_interval.tv_usec = 0;
        tmrv.it_value.tv_sec = 0;
        tmrv.it_value.tv_usec = usec;
        setitimer(ITIMER_REAL, &tmrv, NULL);
        sigpause (SIGVTALRM);
}
#endif






/*
 * Averager functions.
 */

void
av_Zero (Averager *a)
{
	a->a_NSamples = 0;
	a->a_Sum = 0.0;
}


void
av_Accumulate (float v, Averager *a)
{
	if (v != BADVAL)
	{
		a->a_Sum += v;
		(a->a_NSamples)++;
	}
}



float
av_GetAverage (Averager *a)
{
	if (a->a_NSamples > 0)
		return (a->a_Sum/a->a_NSamples);
	return (BADVAL);
}



/*
 * Calibrate generic weather station data.
 */
void calibrate_ws (struct ws_data *ws)
{
/*
 * Validate raw data values, scale.
 */
	if (ws->intemp != BADVAL)
		ws->intemp = ws->m_intemp * (ws->intemp + ws->a_intemp);

	if (ws->outtemp != BADVAL)
		ws->outtemp = ws->m_outtemp * (ws->outtemp + ws->a_outtemp);

	if (ws->pres != BADVAL)
		ws->pres = ws->m_pres * (ws->pres + ws->a_pres);

	if (ws->inhum != BADVAL)
		ws->inhum = ws->m_inhum * (ws->inhum + ws->a_inhum);

	if (ws->outhum != BADVAL)
		ws->outhum = ws->m_outhum * (ws->outhum + ws->a_outhum);

	if (ws->rain != BADVAL)
		ws->rain = ws->m_rain * ws->rain;

	if (ws->wspd != BADVAL)
		ws->wspd =  ws->m_wspd * ws->wspd;

	if (ws->wdir != BADVAL)
		ws->wdir += ws->a_wdir;
		if (ws->wdir >= 360.0) ws->wdir -= 360.0;
		else if (ws->wdir < 0.0) ws->wdir += 360.0;

	return;
}



/*
 * Zero the weather station averages.
 */
void wsd_av_zero (struct ws_data *ws)
{
	av_Zero (&ws->av_outtemp);
	av_Zero (&ws->av_outhum);
	av_Zero (&ws->av_wchill);
	av_Zero (&ws->av_dp);
	av_Zero (&ws->av_pres);
	av_Zero (&ws->av_cpres);
	av_Zero (&ws->av_dpdt);
	av_Zero (&ws->av_wspd);
	av_Zero (&ws->av_intemp);
	av_Zero (&ws->av_inhum);
	av_Zero (&ws->av_uwind);
	av_Zero (&ws->av_vwind);
	ws->peak_gust = 0;
}



/*
 * Sum the weather station data for the averages.
 */
void wsd_av_sum (struct ws_data *ws)
{

/*
 * Check and keep track of each field individually.
 */
	av_Accumulate (ws->outtemp, &ws->av_outtemp);
	av_Accumulate (ws->outhum, &ws->av_outhum);
	av_Accumulate (ws->wchill, &ws->av_wchill);
	av_Accumulate (ws->dp, &ws->av_dp);
	av_Accumulate (ws->pres, &ws->av_pres);
	av_Accumulate (ws->cpres, &ws->av_cpres);
	av_Accumulate (ws->dpdt, &ws->av_dpdt);
	av_Accumulate (ws->wspd, &ws->av_wspd);
	av_Accumulate (ws->intemp, &ws->av_intemp);
	av_Accumulate (ws->inhum, &ws->av_inhum);
/*
 * Save the highest wind speed.
 */
	if (ws->wspd > ws->peak_gust)
		ws->peak_gust = ws->wspd;
/*
 * Convert polar wind speed / direction into cartesian coordinates for summing.
 */
	if (ws->wdir != BADVAL)
	{
		av_Accumulate (ws->wspd*cos (ws->wdir*DEGTORAD),&ws->av_uwind);
		av_Accumulate (ws->wspd*sin (ws->wdir*DEGTORAD),&ws->av_vwind);
	}
}




/*
 * Average the wind speed and direction.
 */
void wsd_average (struct ws_data *ws)
{
	float av_rad_wdir;

/*
 * Convert the cartesian average wind values into average polar direction.
 * Deal with the atan2 output which is PI at 180 deg and -PI at 181 deg.
 */
	av_rad_wdir = (float) atan2 (av_GetAverage (&ws->av_vwind),
			av_GetAverage (&ws->av_uwind));

/*
 * Zero wind speeds mean bad average wind direction.
 * You can't average XY converted direction without the speed component.
 * Fake it and return the most recent wind direction, kind of scientifically
 * ugly, but better for web page output.
 */
	if (av_GetAverage (&ws->av_wspd) <= 0.0)
	{
		if (ws->wdir != BADVAL)
			ws->calc_av_wdir = ws->wdir;
		else
		    ws->calc_av_wdir = BADVAL;
	}
	else
	{
	    if (av_rad_wdir >= 0)
		ws->calc_av_wdir = av_rad_wdir * (180.0 / M_PI);
	    else
		ws->calc_av_wdir = 360.0 - av_rad_wdir * (-180.0 / M_PI);
	}
}


/*
 * Output the weather station sample averages data to a file for use by finger.
 */
void wsd_fingfile(struct ws_data *ws, char *dir)
{
	char tstr[30];
	FILE *fingfp;
	FILE *efingfp;
	char *fname = "finger.txt";
	char *efname = "efinger.txt";
	char filename[130];
	float average1, average2;

/*
 * Open the metric finger data file
 */
	strcpy (filename, dir);
	strcat (filename, fname);

	fingfp = fopen(filename, "w");

	if (fingfp == (FILE *)NULL)
		return;

/*
 * Open the English finger data file
 */
	strcpy (filename, dir);
	strcat (filename, efname);

	efingfp = fopen(filename, "w");

	if (efingfp == (FILE *)NULL)
		return;

	strcpy (tstr, asctime(localtime(&ws->samp_time)));
	tstr[strlen(tstr)-1] = '\0';

	fprintf (fingfp, "Time=%s\n", tstr);
	fprintf (efingfp, "Time=%s\n", tstr);

	average1 = av_GetAverage (&ws->av_outtemp),
	average2 = av_GetAverage (&ws->av_outhum),
	fprintf (fingfp, "Temperature=%.2f Deg C\nHumidity=%.2f %%\n",
			average1, average2);
	fprintf (efingfp, "Temperature=%.2f Deg F\nHumidity=%.2f %%\n",
			(NINEFIFTHS * average1) + 32.0, average2);

	if (!Shortfing)
	{
	    average1 = av_GetAverage (&ws->av_wchill);
	    average2 = av_GetAverage (&ws->av_dp);
	    fprintf (fingfp, "Windchill=%.2f Deg C\nDewpoint=%.2f Deg C\n",
			average1, average2);
	    fprintf (efingfp, "Windchill=%.2f Deg F\nDewpoint=%.2f Deg F\n",
	    (NINEFIFTHS * average1) + 32.0, (NINEFIFTHS * average2) + 32.0);

	    average1 = av_GetAverage (&ws->av_pres);
	    average2 = av_GetAverage (&ws->av_dpdt);
	    fprintf (fingfp, "Pressure=%.2f mB\nDpdt=%.2f mB/H\n",
	    		average1, average2);
	    fprintf (efingfp, "Pressure=%.2f in-Hg\nDpdt=%.2f in-Hg/H\n",
			average1 * MB2IN, average2 * MB2IN);

	    fprintf (fingfp, "Aero");	/* print AeroPres in long finger msg */
	    fprintf (efingfp, "Aero");
	}

	average1 = av_GetAverage (&ws->av_cpres);
	fprintf (fingfp, "Pressure=%.2f mB\n", average1);
	fprintf (efingfp, "Pressure=%.2f in-Hg\n", average1 * MB2IN );

	average1 = av_GetAverage (&ws->av_wspd);
	fprintf (fingfp, "Windspeed=%.2f Km/H\nWinddir=%.1f Deg\n",
			average1, ws->calc_av_wdir);
	fprintf (efingfp, "Windspeed=%.2f Mi/H\nWinddir=%.1f Deg\n",
			average1 * KMH2MIH, ws->calc_av_wdir);

	fprintf (fingfp, "Peakgust=%.2f Km/H\n", ws->peak_gust);
	fprintf (efingfp, "Peakgust=%.2f Mi/H\n", ws->peak_gust * KMH2MIH);

	fprintf (fingfp, "Rain=%.2f mm\n", ws->rain);
	fprintf (efingfp, "Rain=%.2f in\n", ws->rain * MM2IN);

	average1 = av_GetAverage (&ws->av_intemp);
	fprintf (fingfp, "Intemp=%.2f Deg C\n", average1);
	fprintf (efingfp, "Intemp=%.2f Deg F\n",
			(NINEFIFTHS * average1) + 32.0);

	average1 = av_GetAverage (&ws->av_inhum);
	fprintf (fingfp, "Inhum=%.2f %%\n", average1);
	fprintf (efingfp, "Inhum=%.2f %%\n", average1);

	fclose (fingfp);
	fclose (efingfp);
}


/*
 * Output the weather station sample averages data to a text data file.
 */
void wsd_datafile(struct ws_data *ws, char *dir)
{
	FILE *datafp;
	char *fname = "davis.";
	char *fext = ".asc";
	char filename[130];
	char tstr[30];
	struct tm *localt;
	int year;

	localt = localtime(&ws->samp_time);

	if (localt->tm_year > 80) year = 1900 + localt->tm_year;
	else year = 2000 + localt->tm_year;

	sprintf (tstr,"%d%02d%02d",year,localt->tm_mon+1,localt->tm_mday);

	strcpy (filename, dir);
	strcat (filename, fname);
	strcat (filename, tstr);
	strcat (filename, fext);
	/* davis.19970314.dat */

	datafp = fopen(filename, "a");

	if (datafp == (FILE *)NULL)
		return;

	fprintf (datafp, "%ld %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f\n",
		(long)ws->samp_time,
		av_GetAverage (&ws->av_outtemp),
		av_GetAverage (&ws->av_outhum),
		av_GetAverage (&ws->av_wchill),
		av_GetAverage (&ws->av_dp),
		av_GetAverage (&ws->av_pres),
		av_GetAverage (&ws->av_dpdt),
		av_GetAverage (&ws->av_cpres),
		av_GetAverage (&ws->av_wspd),
		ws->calc_av_wdir,
		ws->peak_gust,
		ws->rain,
		av_GetAverage (&ws->av_intemp),
		av_GetAverage (&ws->av_inhum));

	fclose (datafp);
}


/*
 * Calculate all of the derived fields defined in fields.
 */
void calc_derived (struct ws_data *ws, unsigned int fields)
{
	float ews, lews;
	float wspd;
	int min;
	static int last_min=-1, first=1;
	static float prev_pres;

/*
 * Calculate aeronautical pressure correction as per the Smithsinian tables.
 */
    if (fields & DER_CPRES)
    {
	ws->cpres = ws->pres * (float)pow(
			(1.0 + 0.00008423 * (double)(ws->elevation /
			(pow((double)ws->pres, 0.1903)))), 5.255);

	if ((ws->cpres < 600.0) || (ws->cpres > 1400.0)) ws->cpres = BADVAL;
    }
/*
 * Calculate dp/dt over a 15 minute interval scaled to an hour.
 */
    if (fields & DER_DPDT)
    {
        if (first) ws->dpdt = BADVAL;

	min = gmt_minute ();

	if (last_min != min)	/* check once a minute */
	{
		if (min % 15 == 0)
		{
		    if (first)
			first=0;
		    else
			ws->dpdt=(av_GetAverage(&ws->av_pres)-prev_pres)*4.0;

		    prev_pres = av_GetAverage (&ws->av_pres);
		}
	}
	if ((ws->dpdt < -50.0) || (ws->dpdt > 50)) ws->dpdt = BADVAL;

	last_min = min;
    }

/*
 * Calculate the dewpoint.
 */

    if (fields & DER_DEWP)
    {
        if (ws->outhum < BADVAL + 1)	/* correct for NAN case */
	{
	    ws->dp = BADVAL;
	}
	else
	{
	    ews = ws->outhum * 0.01 * (float)exp((double)((17.502*ws->outtemp) /
				(240.9+ws->outtemp)));

	    lews = (float)log((double)ews);

	    ws->dp = (240.9 * lews) / (17.5 - lews);

	    if ((ws->dp < -80.0) || (ws->dp > 140) || (ws->dp > ws->outtemp))
	    	ws->dp = BADVAL;
	}
    }

/*
 * Calculate wind chill according to the Siple method
 */
    if (fields & DER_WCHIL)
    {
	wspd = ws->wspd *  KPH2MPS;

	if ((wspd >= 1.79) && (ws->outtemp < 33.0))
	{
		ws->wchill = 33.0 + (ws->outtemp - 33.0) *
		(0.550 + (0.417 * (float)sqrt((double)wspd)) - 0.0454 * wspd);
	}
	else ws->wchill = ws->outtemp;

	if ((ws->wchill < -80.0) || (ws->wchill > 33.0)) ws->wchill = BADVAL;
    }
}


/*
 * Dump the weather station data structure.
 */
void wsd_dump (struct ws_data *ws)
{
	printf ("\noutside temperature: %3.1f degrees C", ws->outtemp);
	printf ("\ninside temperature: %3.1f degrees C", ws->intemp);
	printf ("\noutside humidity: %3.1f percent", ws->outhum);
	printf ("\ninside humidity: %3.1f percent", ws->inhum);
	printf ("\npressure: %4.2f mb", ws->pres);
	printf ("\nwind speed: %4.1f km/hr", ws->wspd);
	printf ("\nwind direction: %4.1f degrees, from the %s",
				ws->wdir, ws->wdirname);
	printf ("\nrain accumulation: %3.2f mm", ws->rain);
	printf ("\nwind chill: %3.1f degrees c", ws->wchill);
	printf ("\ndewpoint: %3.1f degrees C", ws->dp);
	printf ("\ndp/dt: %4.2f mb/hour", ws->dpdt);
	printf ("\n");
}



/*
 * Dump the weather station average data.
 */
void wsd_av_dump (struct ws_data *ws)
{
	printf ("\naverage temperature: %3.1f degrees C",
			av_GetAverage (&ws->av_outtemp));
	printf ("\naverage humidity: %3.1f percent",
			av_GetAverage (&ws->av_outhum));
	printf ("\naverage wind chill: %3.1f degrees c",
			av_GetAverage (&ws->av_wchill));
	printf ("\naverage dewpoint: %3.1f degrees C",
			av_GetAverage (&ws->av_dp));
	printf ("\naverage pressure: %4.2f mb",
			av_GetAverage (&ws->av_pres));
	printf ("\naverage Calibrated pressure: %4.2f mb",
			av_GetAverage (&ws->av_cpres));
	printf ("\naverage dp/dt: %4.2f mb/hour",
			av_GetAverage (&ws->av_dpdt));
	printf ("\naverage wind speed: %4.1f m/sec",
			av_GetAverage (&ws->av_wspd));
	printf ("\npeak gust: %4.1f m/sec",
			ws->peak_gust);
	printf ("\naverage wind direction: %4.1f degrees",
			ws->calc_av_wdir);
	printf ("\naverage inside temperature: %3.1f degrees C",
			av_GetAverage (&ws->av_intemp));
	printf ("\naverage inside humidity: %3.1f percent",
			av_GetAverage (&ws->av_inhum));
	printf ("\n");
}


/*
 * Read in the lat/lon/elev data from the config file.
 */
void read_config_location (struct ws_data *ws)
{
	ws->latitude = CF_GetConfigFloat ("latitude", DF_LATITUDE);
	ws->longitude = CF_GetConfigFloat ("longitude", DF_LONGITUDE);
	ws->elevation = CF_GetConfigFloat ("elevation", DF_ELEVATION);
	CF_GetConfigString ("location", ws->location, DF_ST_LOCATION);
	CF_GetConfigString ("model", ws->model, DF_ST_MODEL);
}


/*
 * Dump in the lat/lon/elev data from the config file.
 */
void dump_config_location (struct ws_data *ws)
{
	printf ("\nStation Information:");
	printf ("\nlocation:  %s", ws->location);
	printf ("\nmodel:     %s", ws->model);
	printf ("\nlatitude:  %3.3f degrees N", ws->latitude);
	printf ("\nlongitude: %3.3f degrees W", ws->longitude);
	printf ("\nelevation: %5.3f meters", ws->elevation);
	printf ("\n");
}


/*
 * Read in the calibration data from the config file.
 */
void read_config_cals (struct ws_data *ws)
{
	ws->a_outtemp = CF_GetConfigFloat ("out_temp_add", 0.0);
	ws->a_outhum = CF_GetConfigFloat ("out_hum_add", 0.0);
	ws->a_pres = CF_GetConfigFloat ("out_pres_add", 0.0);
	ws->a_wdir = CF_GetConfigFloat ("out_wdir_add", 0.0);
	ws->a_intemp = CF_GetConfigFloat ("in_temp_add", 0.0);
	ws->a_inhum = CF_GetConfigFloat ("in_hum_add", 0.0);

	ws->m_outtemp = CF_GetConfigFloat ("out_temp_mul", 1.0);
	ws->m_outhum = CF_GetConfigFloat ("out_hum_mul", 1.0);
	ws->m_pres = CF_GetConfigFloat ("out_pres_mul", 1.0);
	ws->m_wspd = CF_GetConfigFloat ("out_wspd_mul", 1.0);
	ws->m_rain = CF_GetConfigFloat ("out_raina_mul", 1.0);
	ws->m_intemp = CF_GetConfigFloat ("in_temp_mul", 1.0);
	ws->m_inhum = CF_GetConfigFloat ("in_hum_mul", 1.0);
}


/*
 * Dump the calibration data
 */
void dump_config_cals (struct ws_data *ws)
{
	printf ("\nCalibration Additives:");
	printf ("\n Out Temp:  %3.3f deg C", ws->a_outtemp);
	printf ("\n Out Hum:   %3.3f percent", ws->a_outhum);
	printf ("\n Pressure:  %3.3f mb", ws->a_pres);
	printf ("\n Wind Dir:  %3.3f deg", ws->a_wdir);
	printf ("\n In Temp:   %3.3f deg C", ws->a_intemp);
	printf ("\n In Hum:    %3.3f percent", ws->a_inhum);

	printf ("\n\nCalibration Multipliers:");
	printf ("\n Out Temp:  %3.3f", ws->m_outtemp);
	printf ("\n Out Hum:   %3.3f", ws->m_outhum);
	printf ("\n Pressure:  %3.3f", ws->m_pres);
	printf ("\n Wind Speed:%3.3f", ws->m_wspd);
	printf ("\n Rain:      %3.3f", ws->m_rain);
	printf ("\n In Temp:   %3.3f", ws->m_intemp);
	printf ("\n In Hum:    %3.3f", ws->m_inhum);
	printf ("\n\n");
}

void hex_dump (unsigned char *str, int len)
{
	int i;

	printf ("\nHex Dump: ");

	for (i=0; i<len; i++)
		printf ("%02X ", str[i]);

	printf ("\n");
}

void str_hex_dump (char *rstr, char *lbl, unsigned char *str, int len)
{
	int i;
	char digstr[4];

	sprintf (rstr, "%s Hex: ", lbl);

	for (i=0; i<len; i++)
	{
		sprintf (digstr, "%02X ", str[i]);
		strcat (rstr, digstr);
	}
}

