/*
 *-----------------------------------------------------------------------------
 * dv.c the Davis Data Grabber Daemon May 1, 1997 [fc]
 * (c) 2005 G. Forrest Cook, Jonathan Corbet
 * License: GPL Version 2.0
 * Version 2.0: Revisited Jun 20, 2000 [fc], made standalone
 * Version 2.1: May 16, 2005 [fc] added include <time.h> to WsIngest.c to
 * compile under Slackware 10.0 (thanks to Brad Grant.)
 * Version 2.2: May 17, 2005 [fc] added units to the finger file, corrected
 * README to reflect Km/h wind speeds (thanks to Pascal Grandjean.)
 * Detected a bug, Dec 8, 2005, if outside temps go < 0C, they read as bad.
 * Version 2.3: Feb 18, 2006 [fc] fixed negative outside temperature bug.
 *  Also fixed a couple of compiler warnings with unsigned int and char * casts.
 *  Added -i mode to dump integer testing data
 *-----------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <syslog.h>
#include <time.h>
#include <fcntl.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include "WsIngest.h"
#include "Config.h"
#include "dv.h"
#include "ccitt.h"

int Help=0;
int Exit=0;
int Verb=0;
int Dump=0;
int DumpInt=0;
int Shortfing=0;


void Exception (int s)
{
        printf ("Exception on %d\n", s);
	exit (0);
}

/*
 * davis Daemon start
 */

int main (int argc, char **argv)
{
	int pd, rlen, samp_interval;
	char sio_port[80];
	char log_str[80];
	char configf[200];
	unsigned char ret_buf[600];
	char data_dir[120];
	char finger_dir[120];
	int opterr = 0, cfset = 0, opt;
	struct ws_data ws;
	char gripe[80];
	time_t rtime, ltime=-1;
	int last_time;

/*
 * command line stuff snarfed from daemon.c
 */
	gmt_log ("Davis Daemon V2.3 Starting");

/*
 * Let's look at our options.
 */
	opterr = 0;
	while ((opt = getopt (argc, argv, "c:div")) != EOF)
	{
		switch (opt)
		{
		/*
		 * Oh, goodie, we needed a config file.
		 */
		    case 'c':
			CF_SetConfigFile (optarg, "dv");
			strcpy (configf, optarg);
			cfset = 1;
			break;
		/*
		 * Verbose mode?
		 */
		    case 'v':
			Verb = 1;
			gmt_log ("Verbose Mode");
			break;
		/*
		 * Dump mode?
		 */
		    case 'd':
			Dump = 1;
			gmt_log ("Dump Mode");
			break;
		/*
		 * Dump integer data mode?
		 */
		    case 'i':
			DumpInt = 1;
			gmt_log ("Dump Integer Mode");
			break;
		/*
		 * Oops.
		 */
		    default:
			sprintf (gripe, "Unknown option: %c", opt);
			syslog (LOG_ERR, gripe);
			break;
		}
	}
	if (!cfset)
		CF_SetConfigFile (DF_CF_FILE, "dv");
/*
 * Read the config file options.
 */
	CF_GetConfigString ("dvport", sio_port, DF_DV_PORT);
	CF_GetConfigInt ("samp_interval", &samp_interval, DF_SAMPLE_INTERVAL);
	CF_GetConfigString ("data_dir", data_dir, DF_DATADIR);
	CF_GetConfigString ("finger_dir", finger_dir, DF_FINGDIR);
	CF_GetConfigInt ("short_finger", &Shortfing, DF_FINGMODE);
	read_config_location (&ws);
	read_config_cals (&ws);

/*
 * Print out various startup info.
 */ 
	sprintf (log_str, "Config File:		%s", configf);
	gmt_log (log_str);
	sprintf (log_str, "Serial Port:		%s", sio_port);
	gmt_log (log_str);
	sprintf (log_str, "Sample Interval:	%d", samp_interval);
	gmt_log (log_str);
	sprintf (log_str, "Data Directory:	%s", data_dir);
	gmt_log (log_str);
	sprintf (log_str, "Finger Directory:	%s", finger_dir);
	gmt_log (log_str);

	if (Verb)
	{
		dump_config_location (&ws);
		dump_config_cals (&ws);
	}

/*
 * Open the Davis port.
 */ 
	pd = dvport_setup (sio_port);

/*
 * Initialize the multi-sample data averages.
 */ 
	wsd_av_zero (&ws);

/*
 * Main loop
 */
	while (!Exit)
	{
		/* printf ("\nmain loop\n"); */
		rlen = dv_loop_comd (pd, ret_buf);

		if (rlen)
		{
		    if (grok_davis (rlen, ret_buf, &ws))
		    {
			if (Dump) wsd_dump (&ws);

			time (&rtime);
			ws.snap_time = rtime;

			wsd_av_sum (&ws);

			/*
			 * Output averages every samp_interval.
			 */
			if (rtime % samp_interval < last_time)
			{
			    /*
			     * Save the sample time rounded back to the
			     * nearest sample interval.  Prevent duplicate
			     * time vals which may occur occasionally.
			     */
			    ws.samp_time = rtime - (rtime % samp_interval);

			    if (ws.samp_time == ltime)
				ws.samp_time += samp_interval;

			    ltime = ws.samp_time;

			    /*
			     * Calculate the averages and output them.
			     */
			    wsd_average (&ws);

			    if (Dump) wsd_av_dump (&ws);
			    wsd_datafile (&ws, data_dir);
			    wsd_fingfile (&ws, finger_dir);
			    wsd_av_zero (&ws);
			}
			last_time=rtime % samp_interval;
		    }
		}
	}
	close (pd);

	exit (-1);
}


/*
 * Interpret the data from the davis.
 * Check, scan, validate, metricize the davis DRTDA output. 
 */
int grok_davis (int rlen, unsigned char *ret_buf, struct ws_data *ws)
{
	struct raw_dv_data raw;
	char log_str[80];
	unsigned short crc=0, read_crc;
	int i;
/*
 * Check for a long enough string before scanning it.
 */
	if (rlen < 19)
	{
		sprintf (log_str, "string too short, %d chars", rlen);
		gmt_log (log_str);
		return (-1);
	}

/*
 * Make sure we are synchronized with the start bytes.
 */
	if (ret_buf[0] != 0x06 || ret_buf[1] != 0x01)
	{
		sprintf (log_str, "ACK and SOH not received, len:%d", rlen);
		gmt_log (log_str);
		return (-1);
	}

/*
 * Pick the various pieces of data from the binary string.
 * Deal with below zero outside temperature case specially.
 */

	raw.iintemp = (int)ret_buf[2] + ((int)ret_buf[3]<<8);
	raw.iouttemp = (int)ret_buf[4] + ((int)ret_buf[5]<<8);
	if (raw.iouttemp > 32768) raw.iouttemp -= 65536;
	raw.iwspd = (int)ret_buf[6];
	raw.iwdir = (int)ret_buf[7] + ((int)ret_buf[8]<<8);
	raw.ipres = (int)ret_buf[9] + ((int)ret_buf[10]<<8);
	raw.iinhum = (int)ret_buf[11];
	raw.iouthum = (int)ret_buf[12];
	raw.irain = (int)ret_buf[13] + ((int)ret_buf[14]<<8);
	/* 2 unused bytes */

/*
 * Read the crc data.
 */
	read_crc = (int)ret_buf[17] + ((int)ret_buf[18]<<8);

/*
 * crc for ret_buf[2 to 18] should evolve to 0
 */
	for (i=2; i<=18; i++)
		crc = Crc_table [(crc >> 8) ^ ret_buf[i]] ^ (crc << 8);

/*
 * Quick exit for bad CRC data.
 */
	if (crc != 0)
	{
		sprintf (log_str, "CRC Error, crc:%d, read_crc:%d",
			crc, read_crc);
		gmt_log (log_str);
		return (-1);
	}

/*
 * After all that, we can assume the data is as good as the station
 * can provide.
 */

/*
 * Dump the integer data if requested.
 */ 
	if (DumpInt)
printf ("\nIn_T:%d,Out_T:%d,Wspd:%d,Wdir:%d,Pres:%d,In_H:%d,Out_H:%d,Rain:%d",
		raw.iintemp, raw.iouttemp, raw.iwspd, raw.iwdir,
		raw.ipres, raw.iinhum, raw.iouthum, raw.irain);

/*
 * Convert the davis to metric floats.
 * calibrate the data
 */
	dv_to_metric_float (ws, &raw);
	calibrate_ws (ws);

/*
 * Calculate derived data.
 */
	calc_derived (ws, DER_CPRES|DER_DPDT|DER_DEWP|DER_WCHIL);

	return (1);
}




/*
 * Send a loop command to the davis.
 */
int dv_loop_comd (int pd, unsigned char *ret_buf)
{
# define READMS 250	/* read delay time in milliseconds, 250 at 2400 baud */
# define DAVISDLY 4	/* delay time before read in seconds */

	char log_str[600];
	static int first = 1;
	static unsigned char tmp_buf[200];
	static unsigned char out_buf[10];
	int i, j, k=0, rv, rlen=0;

	ret_buf[0] = '\0';
/*
 * Flush any old stuff from the port (Solaris needs it).
 */
	ioctl(pd, TCFLSH, TCIFLUSH);

/*
 * Establish sync with the davis by sending ?\r until ! or BEL! is returned.
 */
	if (first)
	{
	    do
	    {
		/* clear out the read data */
		do
		{
			rv=dvport_read (pd, tmp_buf, 30, 0, READMS);
		} while (rv);

		dvport_write (pd, (unsigned char *)"?\r", 2);

		tmp_buf[0] = '\0';
		rv=dvport_read (pd, tmp_buf, 10, 1, READMS);

		if (Verb && rv>0)
		{
			str_hex_dump (log_str, "Read ", tmp_buf, rv);
			gmt_log (log_str);
		}

		if (k++ > 18)
		{
			if (rv > 16)
			{
		gmt_log ("In loop mode already? Power Cycle the Davis WS");
			}
			else
		gmt_log ("failed handshake, check Davis RS-232 connection.");
			return (0);
		}
	    }
	    while (tmp_buf[0] != '!');
  
	    if (Verb) gmt_log ("Handshake OK");
	}
	first = 0;

/*
 * Now actually send the loop command to the davis, count is 65536-n.
 */
	strcpy ((char *)out_buf, "LOOP");
	out_buf[4] = 255;	/* low byte */
	out_buf[5] = 255;	/* high byte */
	out_buf[6] = '\r';

	dvport_write (pd, out_buf, 7);

/*
 * Delay a bit while the davis gets its shit together
 * This also frees up the console pushbuttons
 */
	sleep (DAVISDLY);

/*
 * Read back the results with several reads until nothing comes in.
 * Under Linux, The read returns with about 50 chars of data the first time,
 * then several bytes for each subsequent read, even with long read times.
 */
	do
	{
		rv=dvport_read (pd, tmp_buf, 30, 0, READMS);

		j=0;
		for (i=rlen; i<rlen+rv; i++) ret_buf[i] = tmp_buf[j++];
		rlen += rv;
	}
	while (rv);

	if (Verb)
	{
		str_hex_dump (log_str, "Read", tmp_buf, rlen);
		gmt_log (log_str);
	}

	return (rlen);
}



/*
 * Read stuff from the davis.
 * Return number of chars read if successful,
 * -1 for read error or 0 for nuthin' happnin'.
 */
int dvport_read (int pd, unsigned char *buf, int rlen, int sec, int msec)
{
	int rv;
	int numsel;
	struct timeval seltime;
	fd_set Sio_r_fds, rfds;
	char err_str[80];

/*
 * Init the select() params.
 */
	FD_ZERO (&Sio_r_fds);
	FD_SET (pd, &Sio_r_fds);

	rfds = Sio_r_fds;

	seltime.tv_sec=sec;
	seltime.tv_usec=1000*msec;

	numsel = select (pd+1, &rfds, (fd_set *)0, (fd_set *)0, &seltime);

	if (numsel < 0)
	{
		perror ("Read Select");
		exit (1);
	}
	else if (numsel)
	{
		/*printf ("\nnumsel:%d\n", numsel);*/
/*
 * Deal with a read if a read file descriptor is set.
 */
		if (FD_ISSET (pd, &rfds))
		{
			/*printf ("\nselect detects rfds\n");*/
			rv=read (pd, buf, rlen);
			if (rv>0)
			{
				buf[rv]=0;
				return (rv);
			}
			else
			{
				sprintf (err_str,"read error, errno:%d",errno);
				gmt_log (err_str);

				return (-1);
			}
		}
	}
	return (0);
}



/*
 * Write out to the davis via the serial port.
 */
void dvport_write (int pd, unsigned char *str, int len)
{
	char *sptr;
	char log_str[300];

	sptr=(char *)str;

	if (Verb)
	{
		printf ("\n");
		str_hex_dump (log_str, "Write", str, len);
		gmt_log (log_str);
	}

/*
 * Write chars out the port, delay between each char to avoid
 * overrunning the davis input buffer.
 * Waiting for the drain seems to help syncing the upcoming read.
 */
	while (*sptr)
	{
		write (pd, sptr++, 1);
		tcdrain(pd);
#ifdef LINUX
		usleep (200000);
#endif
#ifdef SOLARIS
		sol_usleep (200000);
#endif
	}
}


/*
 * Set up the davis serial i/o port parameters.
 */
int dvport_setup (char *portname)
{
	char log_str[80];
	struct termios tio;
	int dvfd=0;

	dvfd = open (portname, O_RDWR);
	/*printf ("\nfdesc:%d\n", dvfd);*/

	if (dvfd < 0)
	{
		sprintf (log_str, "Error: can't open port: %s", portname);
		gmt_log (log_str);
		exit (1);
	}

	tcgetattr ( dvfd, &tio );
#ifdef LINUX
	cfmakeraw ( &tio );
	cfsetospeed ( &tio, B2400 );
	cfsetispeed ( &tio, B2400 );
#endif
#ifdef SOLARIS
	tio.c_cflag &= ~(CSIZE|PARENB|CSTOPB|CRTSCTS|CBAUD);
	tio.c_cflag |= B2400|CS8|CLOCAL|CREAD|HUPCL;

	tio.c_iflag &= ~(BRKINT|INLCR|ICRNL|ISTRIP|IUCLC|IXON|IXOFF|IMAXBEL);
	tio.c_iflag |= IGNBRK|IGNPAR;

	tio.c_oflag &= ~(OPOST|OLCUC|ONLCR);

	tio.c_lflag &= ~(ISIG|ICANON|XCASE|IEXTEN|ECHO|ECHOE|ECHOK|ECHONL);
	tio.c_lflag |= FLUSHO;
#endif
	tcsetattr ( dvfd, TCSAFLUSH, &tio );

	return (dvfd);
}



/*
 * Interpret the data from the davis.
 * Scan, validate, Metricize the davis DRTDA output. 
 * For scientific purposes, weather data is typically stored as metric info.
 * It may be converted back to English units for later display.
 */
void dv_to_metric_float (struct ws_data *ws, struct raw_dv_data *raw)
{
	static char *wind_drs[]={"BAD", "N", "NNE", "NE", "ENE", "E", "ESE",
					"SE", "SSE", "S", "SSW", "SW",
					"WSW", "W", "WNW", "NW", "NNW"};
	float calc_wdir;
	int ind_wdir;

#ifdef BLAB
	printf ("\niwspd:%d", raw->iwspd);
	printf ("\niwdir:%d", raw->iwdir);
	/*printf ("\niwchill:%d", raw->iwchill);*/
	printf ("\niintemp:%d", raw->iintemp);
	printf ("\niouttemp:%d", raw->iouttemp);
	printf ("\nipres:%d", raw->ipres);
	/*printf ("\nidpdt:%d", raw->idpdt);*/
	printf ("\niinhum:%d", raw->iinhum);
	printf ("\niouthum:%d", raw->iouthum);
	printf ("\nidp:%d", raw->idp);
	printf ("\nirain:%d", raw->irain);
#endif

/*
 * Validate raw data values, scale and convert to metric floats.
 */
	if ((raw->iintemp < -600) || (raw->iintemp > 1500)) ws->intemp = BADVAL;
	else	ws->intemp = FIVENINTHS * (((float)raw->iintemp/10.0)-32.0);

	if ((raw->iouttemp < -600) || (raw->iouttemp > 1500))ws->outtemp=BADVAL;
	else	ws->outtemp = FIVENINTHS * (((float)raw->iouttemp/10.0)-32.0);

	if ((raw->ipres < 20000) || (raw->ipres > 40000)) ws->pres = BADVAL;
	else	ws->pres = ((float)raw->ipres * IHG2MBC) / 10.0;

	if ((raw->iinhum < 0) || (raw->iinhum > 100)) ws->inhum = BADVAL;
	else	ws->inhum = (float)raw->iinhum;

	if ((raw->iouthum < 0) || (raw->iouthum > 100)) ws->outhum = BADVAL;
	else	ws->outhum = (float)raw->iouthum;

	if ((raw->irain < 0) || (raw->irain > 9999)) ws->rain = BADVAL;
	else	ws->rain = (float)raw->irain *IN2MMC;

	if ((raw->iwspd < 0) || (raw->iwspd > 255)) ws->wspd = BADVAL;
	else	ws->wspd =  raw->iwspd * MIH2KMH;

	if ((raw->iwdir < 0) || (raw->iwdir > 360))
	{
		ws->wdir = BADVAL;
		strcpy (ws->wdirname, wind_drs[0]);
	}
	else
	{
		ws->wdir = (float)(raw->iwdir);

		calc_wdir = ((float)raw->iwdir + 11.25);
		if (calc_wdir > 360.0) calc_wdir -= 360.0;
		ind_wdir = (int)(calc_wdir / 22.5) + 1;

		/*printf ("\nwdir:%d calc_wdir:%f ind_wdir:%d\n",
			raw->iwdir, calc_wdir, ind_wdir);*/

		strcpy(ws->wdirname,wind_drs[ind_wdir]);
	}

	return;
}
